#include <iostream>
#include <iomanip>
#include <chrono>
#include <cstring>
#include <asio.hpp>
#include <regex>
#include <limits>

const int NTP_PORT = 123;
const int NTP_PACKET_SIZE = 48;

void queryNTPServer(const std::string& server, int utc_offset) {
    try {
        asio::io_context io_context;

        asio::ip::udp::resolver resolver(io_context);
        asio::ip::udp::resolver::query query(asio::ip::udp::v4(), server, std::to_string(NTP_PORT));
        asio::ip::udp::endpoint endpoint = *resolver.resolve(query).begin();

        asio::ip::udp::socket socket(io_context);
        socket.open(asio::ip::udp::v4());

        std::array<unsigned char, NTP_PACKET_SIZE> request{};
        request[0] = 0b11100011;

        auto start_time = std::chrono::high_resolution_clock::now();

        socket.send_to(asio::buffer(request), endpoint);

        std::array<unsigned char, NTP_PACKET_SIZE> response{};
        asio::ip::udp::endpoint sender_endpoint;
        socket.receive_from(asio::buffer(response), sender_endpoint);

        auto end_time = std::chrono::high_resolution_clock::now();
        std::chrono::duration<double> ping = end_time - start_time;

        const unsigned long long NTP_TIMESTAMP_DELTA = 2208988800ULL;
        unsigned long long seconds = (static_cast<unsigned long long>(response[40]) << 24) |
        (static_cast<unsigned long long>(response[41]) << 16) |
        (static_cast<unsigned long long>(response[42]) << 8) |
        (static_cast<unsigned long long>(response[43]));
        seconds -= NTP_TIMESTAMP_DELTA;

        std::time_t raw_time = static_cast<std::time_t>(seconds);
        std::tm* utc_time = std::gmtime(&raw_time);

        std::time_t adjusted_time = raw_time + (utc_offset * 3600);
        std::tm* local_time = std::gmtime(&adjusted_time);

        std::cout << "Details of " << server << "\n";

        if (utc_offset == 0) {
            std::cout << "Time: " << std::put_time(utc_time, "%H:%M:%S") << " UTC\n";
        } else {
            std::cout << "Time (utc" << (utc_offset >= 0 ? "+" : "") << utc_offset << "): "
            << std::put_time(local_time, "%H:%M:%S") << "\n";
        }

        std::cout << "Date: " << std::put_time(local_time, "%Y-%m-%d") << "\n";
        std::cout << "Ping: " << std::fixed << std::setprecision(3) << ping.count() * 1000 << " ms\n";

    } catch (const std::exception& e) {
        std::cerr << "Error: " << e.what() << "\n";
    }
}

int parseUTCOffset(const std::string& timezone) {
    std::regex tz_regex(R"(^utc([+-]?\d+)$)", std::regex::icase);
    std::smatch match;
    if (std::regex_match(timezone, match, tz_regex)) {
        try {
            int offset = std::stoi(match[1]);
            if (offset < -12 || offset > 14) {
                throw std::out_of_range("UTC offset out of range (-12 to +14)");
            }
            return offset;
        } catch (const std::exception&) {
            throw std::invalid_argument("Invalid UTC offset");
        }
    }
    throw std::invalid_argument("Invalid timezone format");
}

int main(int argc, char* argv[]) {
    if (argc < 2 || argc > 3) {
        std::cerr << "Usage: " << argv[0] << " [-t=utcX] <NTP server>\n";
        return 1;
    }

    std::string timezone_flag = "-t=utc0";
    std::string server;

    if (argc == 3) {
        timezone_flag = argv[1];
        server = argv[2];
    } else {
        server = argv[1];
    }

    int utc_offset = 0;
    if (timezone_flag.rfind("-t=", 0) == 0 || timezone_flag.rfind("--timezone=", 0) == 0) {
        std::string timezone = timezone_flag.substr(timezone_flag.find('=') + 1);
        try {
            utc_offset = parseUTCOffset(timezone);
        } catch (const std::exception& e) {
            std::cerr << "Error: " << e.what() << "\n";
            return 1;
        }
    }

    queryNTPServer(server, utc_offset);
    return 0;
}
